/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  gen.h
///  \brief abstract classes for generation of discrete structures
///
///
/// Copyright (C) 2010 Ralf Gugisch
///	Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///	http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef GEN_H
#define GEN_H

#include <limits.h>
#include "latexWriter.hpp"

////////////////////////////////////////////////////////////////////////////////
/// \class generator
/// \brief abstract class for generating structures in backtrack manner.
/// 
/// Iterating through a backtrack tree of depth n=depth() (an abstract
/// virtual function). The m_root is at level -1, leafs are at level
/// n-1, thus indeed we have n+1 levels in the tree. The iteration
/// starts with first son of the m_root (a node on level 0), walks in
/// depth-first manner through the tree, reaching each leaf (the
/// complete structures to generate). The iteration is finished, if
/// the m_root is reached (at level -1).
///
/// Subtrees may be skipped, by jumping to the next node on a level<=lvl by 
///  <tt> gen.next(lvl) </tt>
/// 
/// Iteration may be restarted for all sons of actual node on level lvl by 
/// <tt> gen.begin(lvl) </tt>.
/// (default value lvl=-1, i.e. iterate through all below the m_root). 
/// End of iteration is recognized, whenever a node on level <= lvl is reached.
///  <tt>  for(gen.begin(lvl); !gen.end(lvl); ++gen) ... </tt>
///
/// 
/// T need to have a standard constructor. The constructed structure is
/// saved in a class element sol of type T. 
///
/// Example:
/// 
/// \code
///     generator gen;
///     int       lvl;
///     gen.init(...);
///     for(gen.begin(); !gen.end(); gen.next(lvl)) {
///       lvl=gen.level();
///       if(!test1(*gen,lvl)) continue; 
///       if(!test2(*gen,lvl)) continue; 
///       // ...
///       if(!gen.complete()) { ++lvl; continue; }
///       cout << *gen;
///     }
/// \endcode
/// 
///  The functionality is given in a derived class by overwriting the following
///  protected functions:
/// 
///  - int  depth() const\n
///    denote the depth of the tree (number of levels).  you m_can
///    specify a flat generation  (ie. without backtracking) with depth=1.
///
///  - bool gen_first(int lvl)\n
///    Get first son, current node is on level lvl-1, the new node (the
///    son) will be on level lvl.  If there is no son, return false.
///
///  - bool gen_next(int lvl)\n
///    Get next node on level lvl; current node is on a level >= lvl,
///    thus the new node is either a brother of the current node, or one
///    of an ancestor thereof. If there is no brother on the given
///    level, return false.
/// 
////////////////////////////////////////////////////////////////////////////////

template<class T> class generator {
private: 
  int m_lvl; ///< Level of actual node; m_root is on level -1, leafs have depth()-1
protected:
  T m_sol; ///< The (maybe yet incomplete) solution, initialized by derived class.

  /// get first son of current node (on level lvl). Ie. initialize part
  /// of level lvl in sol in first possible way.
  virtual bool gen_first(int lvl)=0;  
  
  /// Get next node on level lvl. Ie. fill part of level lvl in sol in
  /// the next possible way.
  virtual bool gen_next(int lvl) =0;

  /// Denote depth of tree (ie. leafs are on level depth()-1).
  /// Default: No backtrack generation, thus all nodes are on level 0
  virtual int  depth() const                                       { return 1; }

  LATEXCLASS m_latex;
  
public:
  generator(const T& s) : m_sol(s)                          
  { 
	  reset();
	  m_latex.newNode(this->m_sol, "Root");
  }
  virtual ~generator() 
  {}
  
  void writeNode(const std::string& text) { m_latex.newNode(m_sol , text); m_latex.finish_level(false);}

  /// Restart iteration through all nodes on level lvl, i.e.go to the first son 
  /// of the active active node on level lvl.
  /// Default lvl=-1 starts iteration at all, ie. initializes with the first son
  /// of m_root (which is at level -1).
  generator<T>& begin(int lvl=-1)                   { return next(lvl+1,true); }

  /// Proceed to next node in backtrack-tree.
  generator<T>& operator++ ()                      { return next(INT_MAX,false); }

  /// Skip to next node on a level <= lvl.
  /// This way, you m_can specify learn effects/prunings in the backtrack-tree
  /// If restart=true, go back to active node on level lvl-1, and select first son
  /// thereof
  generator<T>& next(int lvl=INT_MAX,bool restart=false)  { 
    bool b=false;
    int d=depth();
    if(d<=0) { m_lvl=-1; return *this; } // RG20061006: Special case depth=0
    if(lvl>=d) lvl=d-1;
    if(restart && m_lvl>=lvl) m_lvl=lvl-1;
    if(m_lvl<lvl) {
      b=gen_first(++m_lvl); 
     m_latex.finish_level(b);
    }
    else 
    {
    	m_latex.finish_level(false);
    	m_lvl=lvl+1; // ie. call gen_next(lvl) in next line.
    }
    while(!b && --m_lvl>=0) {
      b=gen_next(m_lvl);
      m_latex.finish_level(b);
    }
    if(m_lvl != -1)
    {
    	m_latex.newNode(this->m_sol, "");
    }
    return *this;   
  }

  /// reset generation to state before begin():
  generator<T>& reset()                                  { m_lvl=-1; return *this; }
  
  /// Return level of actual node
  int level() const                                                { return m_lvl; }

  /// Is actual node a leaf of bt-tree?
  bool complete() const                                 { return m_lvl>=depth()-1; }

  /// Did we finish with whole bt-tree?
  bool end(int lvl=-1) const                                  { return m_lvl<=lvl; }

  /// Access to the current (maybe still incomplete) solution
  T& operator*()                                                { return  m_sol; }
  /// Access to the current (maybe still incomplete) solution
  T* operator->()                                               { return &m_sol; }
  /// Access to the current (maybe still incomplete) solution
  T const& operator*()  const                                   { return  m_sol; }
  /// Access to the current (maybe still incomplete) solution
  T const* operator->() const                                   { return &m_sol; }

  /// depth of backtrack-tree. Ie. Leafs are on level d-1 (m_root is on level -1)
  int backtrack_depth() const                                { return depth(); }

  /// internal call during backtrack (for building derived backtracks)
  bool backtrack_gen_first(int lvl)                 { return gen_first(m_lvl=lvl); }

  /// internal call during backtrack (for building derived backtracks)
  bool backtrack_gen_next(int lvl)                   { return gen_next(m_lvl=lvl); }
//
//  int get_level(void) const {return m_lvl; }

};

#endif // GEN_H
