/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  partition.cpp
///  \brief Implements ordered partitions of lenght n.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "partition.h"
#include "helpfunction.hpp"
#include <fstream>
#include <cstdlib>

OrdPartition::OrdPartition(int cols, const char* const filename)
{
	std::ifstream in;
	in.open(filename);

	if(!in) {
		std::cerr << "Unable to open input file." << std::endl;
		exit(1);
	}

	resize(cols);
	for(int i=0; i<cols; ++i)
	{
		in >> (*this)[i];
	}
}

int OrdPartition::comp(const OrdPartition& p, int lvl)
{
	if(size() != p.size())
	{
		std::cerr<<"Can't compare partitions of different length"<<std::endl;
		exit(1);
	}
	
	for(std::vector<int>::size_type i=lvl+1; i<p.size(); ++i)
	{
		if((*this)[i] == 1 && p[i] == 0)
		{
			return 1; // this < p 
		}
		if((*this)[i] == 0 && p[i] == 1)
		{
			return -1; // this > p 
		}
	}
	
	return 0; // this == p 
}

bool OrdPartition::scan(std::istream& in, unsigned int n)
{
	// scan for the blocks written by  [ , , , ... ]
	
	clear();
	int x;
	
	in>>std::ws;
	if(in.peek()!='[')
	{
		return false;
	}
	in.ignore(); in>>std::ws;
	while(in && in.peek()!=']') 
	{
		in >> x >> std::ws;

		if(in.peek()==',') 
		{
			in.ignore(); 
			in>>std::ws;
		}
		push_back(x);
	}
	in.ignore(); // removes ']'
	
	if(n!=size()) 
	{
		return false;
	}
	return true;
}

std::ostream & operator<< (std::ostream & os, const OrdPartition& p)
{
	if(p.size() == 0)
	{
		return os<<"[ ]"<<std::endl;
	}
	
	os<<"[ ";
	for(std::vector<int>::size_type i=0; i<p.size()-1; ++i)
	{
		os<<p[i]<<", ";
	}
	os<<p[p.size()-1]<<"]";
	
	return os;
}

std::ostream& OrdPartition::printBlocksizes(std::ostream& os) const
{
	return help::print(std::cout, getBlocksizes(), " ");
}

std::vector<unsigned int> OrdPartition::getBlocksizes(void) const
{
	OrdPartition::const_iterator iter(this->begin()+1);
	OrdPartition::const_iterator iterEnd(this->end());
	std::vector<unsigned int> res;
	unsigned int size =1;
	for(; iter<iterEnd; ++iter)
	{
		if(*iter)
		{
			res.push_back(size);
			size =0;
		}
		++size;
	}
	res.push_back(size);
	return res;
}

std::vector<unsigned int>& UnordPartition::getBiggerSets(unsigned int rank, unsigned int setsize, std::vector<unsigned int>& sets) const
{
	sets.clear();
	
	std::vector<unsigned int> count(rank, 0);

	for(unsigned int i=0; i< rank; ++i)
	{
	    ++count[m_index[i]];
	}
	
	for(unsigned int i=0; i<rank; ++i)
	{
		if(count[i]>setsize)
		{
			sets.push_back(i);
		}
	}
	
	return sets;
}

UnordPartition::UnordPartition(int n)
{
	for(int i=0; i<n; ++i)
	{
		m_index.push_back(i);
	}
}

int UnordPartition::countSets(void) const
{
	std::vector<int> sets(m_index.size(), 0);
	int answer = 0;

	for(std::vector<int>::size_type i=0; i<m_index.size(); ++i)
	{
		sets[m_index[i]] = 1;
	}
	
	for(std::vector<int>::size_type i=0; i<m_index.size(); ++i)
	{
		if(sets[i]>0)
		{
		    ++answer;
		}
	}
	
	return answer;
}

std::ostream & operator<< (std::ostream & os, const UnordPartition& p)
{
	for(std::vector<int>::size_type i=0; i<p.m_index.size(); ++i)
	{
		os<<i<<"->"<<p.m_index[i];
		os<<std::endl;
	}
	
	return os;
}
